Require Export lattice_def.
Require Export lat_prop.
Require Export Shared.
Require Export CompleteLattice.

Global Instance Decidable_impl : forall A B : Prop,
    Decidable A → Decidable B → Decidable (A → B).
  introv (da&Ha) (db&Hb).
  destruct da; destruct db; fold_bool; rew_refl in *;
    ((apply decidable_make with true; solve [ fold_bool; rew_refl* ]) ||
     (apply decidable_make with false; solve [ fold_bool; rew_refl* ])).
Defined.

Global Instance Decidable_EquivDec : forall A (E : EquivDec.t A) x y,
    Decidable (EquivDec.eq x y).
  introv. apply sumbool_decidable. apply EquivDec.dec.
Defined.

Global Instance Decidable_PosetDec : forall A (E : PosetDec.t A) x y,
    Decidable (PosetDec.order x y).
  introv. apply sumbool_decidable. apply PosetDec.dec.
Defined.


Notation "'ifd' b 'then' v1 'else' v2" :=
  (match b with
   | left _ => v1
   | right _ => v2
   end)
  (at level 200, right associativity) : type_scope.


Section ToBeSorted.

Require Export Containers.Maps.

Variables K V : Type.

Section EquivDec_eq.

Hypothesis dec : ∀ x y : V, {x = y} + {x ≠ y}.

Definition EquivDec_eq : EquivDec.t V.
  apply EquivDec.Make with (eq := fun x y => x = y); intros; substs~.
Defined.

Lemma join_trans_1 : forall A (H : PosetDec.t A) (t : JoinDec.t A) (x y z : A),
  z ⊑♯ x → z ⊑♯ (x ⊔♯ y).
Proof. introv S. applys PosetDec.trans S. apply JoinDec.bound1. Qed.

Lemma join_trans_2 : forall A (H : PosetDec.t A) (t : JoinDec.t A) (x y z : A),
  z ⊑♯ y → z ⊑♯ (x ⊔♯ y).
Proof. introv S. applys PosetDec.trans S. apply JoinDec.bound2. Qed.

Lemma meet_same : forall A (H : PosetDec.t A) (t : MeetDec.t A) (x : A),
  x =♯ (x ⊓♯ x).
Proof. introv. apply PosetDec.antisym; auto*. Qed.

End EquivDec_eq.

Lemma Comparable_EquivDec : Comparable V → EquivDec.t V.
  introv C. apply EquivDec_eq. introv. apply decidable_sumbool. apply C.
Defined.

Lemma Comparable_EquivDec_eq : forall H (v1 v2 : V),
  @EquivDec.eq V (Comparable_EquivDec H) v1 v2 → v1 = v2.
Proof. simpls~. Qed.

Hypothesis OK : OrderedType K.
Hypothesis LV : LatticeDec.t V.
Hypothesis TV : TopDec.t V.

Lemma empty_find : forall (m : Map [K, V]) k,
  Empty m → m [k] = None.
Proof.
  introv E.
  apply MapFacts.not_find_in_iff.
  intro I. apply MF.elements_in_iff in I. lets (e&He): (rm I).
  apply MapFacts.elements_Empty in E. rewrite E in He. inverts He.
Qed.

Lemma fold_empty : forall A (f : K → V → A → A) i,
  fold f [] i = i.
Proof. introv. apply~ MapFacts.fold_Empty. apply empty_1. Qed.

Lemma inf_bot : forall x y : V,
  x =♯ (⊥♯) →
  y ⊑♯ x →
  y =♯ (⊥♯).
Proof.
  introv E O. apply PosetDec.antisym.
   refine (PosetDec.trans _ _ _ O _). apply PosetDec.refl; auto*.
   apply BotDec.prop.
Qed.

Lemma join_bot : forall x y z : V,
  x =♯ (⊥♯) →
  y =♯ (⊥♯) →
  z =♯ (⊥♯) →
  x ⊔♯ y =♯ y.
Proof.
  introv Ex Ey Ez. apply PosetDec.antisym.
   apply JoinDec.least_upper_bound; auto*.
  apply PosetDec.trans with (⊥♯); auto*.
Qed.

Lemma Gamma_eq : forall (a A : Type) (La : Lattice.t a) (LA : LatticeDec.t A) (op : A → a) N1 N2,
  Gamma.t _ _ op →
  N1 =♯ N2 → Equiv.eq (op N1) (op N2).
Proof. introv G E. apply Poset.antisym; apply~ Gamma.monotone. Qed.

Variable C : Type.
Variable gamma : V → C → Prop.
Hypothesis G : Gamma.t (℘ C) V gamma.

Lemma gamma_eq : forall x y c,
  x =♯ y →
  gamma x c →
  gamma y c.
Proof. introv E Hg. applys* @Gamma.monotone G Hg. Qed.

Lemma join_not_bot : forall x y : V,
  ¬x =♯ (⊥♯) →
  ¬y =♯ (⊥♯) →
  ¬x ⊔♯ y =♯ (⊥♯).
Proof.
  introv Nx Ny M. asserts F: (x ⊑♯ ⊥♯).
    eapply PosetDec.trans. exact (JoinDec.bound1 x y).
    auto*.
  auto*.
Qed.


Section Flat.

Variable T : Type.
Hypothesis ET : EquivDec.t T.

Global Instance FlatLatTop : TopDec.t (@FlatLatticeDec.TB T).
  apply TopDec.Make with (elem := FlatLatticeDec.TB_Top).
  intro x. destruct x; simpls*.
Defined.

Global Instance FlatLatBot : BotDec.t (@FlatLatticeDec.TB T).
  apply BotDec.Make with (elem := FlatLatticeDec.TB_Bot).
  intro x. destruct x; simpls*.
Defined.

Hypothesis CT : Comparable T.

Global Instance FlatLatComparable : Comparable (@FlatLatticeDec.TB T).
  apply make_comparable. introv.
  destruct x; destruct y;
   try (applys decidable_make false; fold_bool; rew_refl; discriminate);
   try (applys decidable_make true; fold_bool; rew_refl~; fail).
  destruct (@comparable _ CT a a0) as [b E]. destruct b; fold_bool; rew_refl in E.
   substs. applys decidable_make true. fold_bool. rew_refl~.
   applys decidable_make false. fold_bool. rew_refl~. introv A. apply E. inverts~ A.
Defined.

End Flat.

Require Import sign.

Lemma sem_add_monotone : forall v1 v'1 v2 v'2,
  v1 ⊑♯ v'1 → v2 ⊑♯ v'2 →
  Sign.sem_add v1 v2 ⊑♯ Sign.sem_add v'1 v'2.
Proof.
  introv O1 O2.
  destruct v1; destruct v'1; inverts O1;
   destruct v2; destruct v'2; inverts O2;
   simpl; constructors~.
Qed.

End ToBeSorted.


Section FlatLat.

Variable C : Type.
Hypothesis E : EquivDec.t C.

Definition flatGamma (a : @FlatLatticeDec.TB C) (c : C) : Prop :=
  match a with
  | FlatLatticeDec.TB_Top => True
  | FlatLatticeDec.TB_Bot => False
  | FlatLatticeDec.TB_Elem c' => c =♯ c'
  end.

Global Instance FlatGamma : Gamma.t (℘ C) (@FlatLatticeDec.TB C) flatGamma.
  apply Gamma.Make.
   intros [| |c1] [| |c2] I c G; simpls; tryfalse; auto*.
   intros [| |c1] [| |c2] c (H1&H2); simpls; tryfalse; auto*.
    repeat cases_if~. false* H0.
Qed.

End FlatLat.


Module UnitLattice.

Section UnitLattice.

Definition t := unit.

Global Instance Equiv : EquivDec.t t.
  apply EquivDec.Make with (eq := fun _ _ => True); intros; auto*.
Defined.

Global Instance Poset : PosetDec.t t.
  apply PosetDec.Make with (eq := _) (order := fun _ _ => True); intros; auto*.
Defined.

Global Instance Lat : LatticeDec.t t.
  apply LatticeDec.Make with (porder := Poset).
   apply JoinDec.Make with (op := fun _ _ => tt); simpls*.
   apply MeetDec.Make with (op := fun _ _ => tt); simpls*.
   apply BotDec.Make with (elem := tt); simpls*.
   exact (fun _ _ => tt).
Defined.

Global Instance Top : TopDec.t t.
  apply TopDec.Make with (elem := tt); simpls*.
Defined.

Definition gamma : t → unit → Prop := fun _ _ => True.

Global Instance Gamma : Gamma.t (℘ unit) t gamma.
Proof.
  apply Gamma.Make; unfolds gamma; simpl; auto*.
Qed.

End UnitLattice.

End UnitLattice.


Module LiftTop.

Section LiftTop.

Variable A : Type.
Hypothesis L : LatticeDec.t A.

Definition t := option A.

Global Instance Poset : PosetDec.t t.
  apply PosetDec.Make with (eq := LiftEquivDec _)
                           (order := fun x y =>
                                       match x, y with
                                         | Some x, Some y => x ⊑♯ y
                                         | _, None => True
                                         | None, Some _ => False
                                       end);
    repeat intros [?|]; intros; simpl; auto~; tryfalse.
   apply @PosetDec.trans with a0; auto*.
   apply~ @PosetDec.dec.
Defined.

Definition Lat : LatticeDec.t t.
  apply LatticeDec.Make with (porder := Poset).
   apply JoinDec.Make with (op := fun x y =>
                             match x, y with
                             | Some x, Some y => Some (x ⊔♯ y)
                             | _, _ => None
                             end);
    repeat intros [?|]; intros; simpl; auto~.
   apply MeetDec.Make with (op := fun x y =>
                             match x, y with
                             | Some x, Some y => Some (x ⊓♯ y)
                             | None, _ => y
                             | _, None => x
                             end);
    repeat intros [?|]; intros; simpl; auto~.
   apply BotDec.Make with (elem := Some (⊥♯)); intros [?|]; simpls*.
   exact (fun x y => y).
Defined.

Definition Top : TopDec.t t.
  apply TopDec.Make with None.
  intros [?|]; simpls~.
Defined.

Variable C : Type.
Variable gamma0 : A → C → Prop.
Hypothesis G : Gamma.t (℘ C) A gamma0.

Definition gamma t c :=
  match t with
  | None => True
  | Some t => gamma0 t c
  end.

Global Instance Gamma : @Gamma.t (℘ C) t _ Lat gamma.
Proof.
  apply Gamma.Make.
   intros [?|] [?|] I c Hg; simpl; auto~.
    applys* @Gamma.monotone G Hg.
    inverts I.
   intros [?|] [?|] c [Hg1 Hg2]; simpls; auto~.
    applys~ @Gamma.meet_morph G. simpls*.
Qed.

End LiftTop.

End LiftTop.


Module LiftBot.

Section LiftBot.

Variable A : Type.
Hypothesis L : LatticeDec.t A.

Definition t := option A.

Global Instance Poset : PosetDec.t t.
  apply PosetDec.Make with (eq := LiftEquivDec _)
                           (order := fun x y =>
                                       match x, y with
                                         | Some x, Some y => x ⊑♯ y
                                         | None, _ => True
                                         | Some _, None => False
                                       end);
    repeat intros [?|]; intros; simpl; auto~; tryfalse.
   apply @PosetDec.trans with a0; auto*.
   apply~ @PosetDec.dec.
Defined.

Definition Lat : LatticeDec.t t.
  apply LatticeDec.Make with (porder := Poset).
   apply JoinDec.Make with (op := fun x y =>
                             match x, y with
                             | Some x, Some y => Some (x ⊔♯ y)
                             | None, _ => y
                             | _, None => x
                             end);
    repeat intros [?|]; intros; simpl; auto~.
   apply MeetDec.Make with (op := fun x y =>
                             match x, y with
                             | Some x, Some y => Some (x ⊓♯ y)
                             | _, _ => None
                             end);
    repeat intros [?|]; intros; simpl; auto~.
   apply BotDec.Make with (elem := None); intros [?|]; simpls*.
   exact (fun x y => y).
Defined.

Hypothesis T : TopDec.t A.

Definition Top : TopDec.t t.
  apply TopDec.Make with (Some (⊤♯)).
  intros [?|]; simpls~. apply T.
Defined.

Variable C : Type.
Variable gamma0 : A → C → Prop.
Hypothesis G : Gamma.t (℘ C) A gamma0.

Definition gamma t c :=
  match t with
  | None => False
  | Some t => gamma0 t c
  end.

Global Instance Gamma : @Gamma.t (℘ C) t _ Lat gamma.
Proof.
  apply Gamma.Make.
   intros [?|] [?|] I c Hg; simpl; auto~.
    applys* @Gamma.monotone G Hg.
    inverts Hg.
   intros [?|] [?|] c [Hg1 Hg2]; simpls; auto~.
    applys~ @Gamma.meet_morph G. simpls*.
Qed.

End LiftBot.

End LiftBot.


Module CoLattice.

Section CoLattice.

Variable T : Type.

Section Poset.

Hypothesis P : PosetDec.t T.

Definition Poset : PosetDec.t T.
  apply PosetDec.Make with (eq := _) (order := fun x y => y ⊑♯ x); auto~.
   auto*.
   introv. apply (PosetDec.dec y x).
Defined.

Hypothesis V : BotDec.t T.

Definition Top : @TopDec.t T Poset.
  apply TopDec.Make with (elem := ⊥♯). simpl. auto*.
Defined.

End Poset.

Hypothesis L : LatticeDec.t T.
Hypothesis TL : TopDec.t T.

Definition Lat : LatticeDec.t T.
  apply LatticeDec.Make with (Poset _).
   apply JoinDec.Make with (fun x y => x ⊓♯ y); simpl; auto*.
   apply MeetDec.Make with (fun x y => x ⊔♯ y); simpl; auto*.
   apply BotDec.Make with (⊤♯). apply TL.
   (* Widen *)
   exact (fun x y => y).
Defined.

End CoLattice.

End CoLattice.


Module MapDefault.

Section MapDefault.

Variables K V : Type.

Require Export Containers.Maps.

Hypothesis OK : OrderedType K.
Hypothesis LV : LatticeDec.t V.
Hypothesis TV : TopDec.t V.

Definition t := (V * (Map [K, {v : V | ¬v =♯ (⊥♯)}]))%type.

Definition read (h : t) k :=
  let (d, m) := h in
  match m [k] with
  | Some v => proj1_sig v
  | None => d
  end.

Definition write (h : t) k v : t :=
  let (d, m) := h in
  match EquivDec.dec v (⊥♯) with
  | right p => (d, m [k <- exist _ v p])
  | left _ => (⊥♯, []) (* Propagating bottom elements *)
  end.

Lemma read_write_eq : forall (h : t) k k' v,
  k === k' →
  read (write h k v) k' =♯ v.
Proof.
  introv E. unfolds read. unfolds write. destruct h as [d m].
  destruct (EquivDec.dec v (⊥♯)).
   rewrite~ empty_find. apply empty_1.
   rewrite MF.add_eq_o; simpls~.
Qed.

Lemma read_write_neq : forall (h : t) k1 k2 v,
  ~ (k1 === k2) ->
  ~ v =♯ (⊥♯) ->
  read (write h k1 v) k2 =♯ read h k2.
Proof.
  introv D NB. unfolds read. unfolds write. destruct h as [d m].
  destruct (EquivDec.dec v (⊥♯)).
   false.
   rewrite MF.add_neq_o; simpls~.
Qed.

Definition mapDefEq f g : Prop :=
  forall k, read f k =♯ read g k.

Definition mapDefLe f g : Prop :=
  forall k, read f k ⊑♯ read g k.

Definition mapDefLeb (f g : t) : bool :=
  let (df, mf) := f in let (dg, mg) := g in
  let check k _ b :=
    b && decide (read f k ⊑♯ read g k) in
  decide (df ⊑♯ dg) &&
  fold check mf true &&
  fold check mg true.

Lemma mapDefLeb_correct : forall f g,
  istrue (mapDefLeb f g) <-> mapDefLe f g.
Proof.
  intros (df&mf) (dg&mg). iff L.
   skip.
   unfolds. rewrite <- and_andb.
    asserts Od: (df ⊑♯ dg).
      skip. (* ≃ K is infinite. *)
    rew_refl. repeat splits~.
     asserts~ Hmf: (forall k v, mf [k] = Some v -> mf [k] = Some v).
      gen L Hmf. generalize mf at - 3. introv.
      apply~ MapFacts.fold_rec. intros. rew_refl. splits.
       skip.
       unfolds read. lets Hmfk: Hmf k.
      clear H2.
Admitted. (* TODO *)

Definition join : t → t → t :=
  fun f g =>
  let (df, mf) := f in let (dg, mg) := g in
  let update f k y h := write h k (read f k ⊔♯ proj1_sig y) in
  fold (update f) mg
    (fold (update g) mf (df ⊔♯ dg, [])).

Definition meet : t → t → t :=
  fun f g =>
  let (df, mf) := f in let (dg, mg) := g in
  let update f k y h := write h k (read f k ⊓♯ proj1_sig y) in
  fold (update f) mg
    (fold (update g) mf (df ⊓♯ dg, [])).

Global Instance Equiv : EquivDec.t t.
  apply EquivDec.Make with mapDefEq.
  firstorder.
  firstorder.
  introv E1 E2. intro k. apply (EquivDec.trans _ _ _ (E1 k) (E2 k)).
  intros f g. set_eq <- b: (mapDefLeb f g && mapDefLeb g f).
   destruct b; fold_bool; rew_refl in EQb.
    lets (Le1&Le2): (rm EQb). apply mapDefLeb_correct in Le1. apply mapDefLeb_correct in Le2.
     left. intro k. apply PosetDec.antisym; [ apply Le1 | apply Le2 ].
    right. rew_logic in *. destruct EQb as [Le|Le]; intro M; apply Le;
      apply mapDefLeb_correct; intro k.
     applys @PosetDec.refl M.
     forwards M': (rm M) k. apply EquivDec.sym in M'. applys @PosetDec.refl M'.
Defined.

Global Instance Poset : PosetDec.t t.
  apply PosetDec.Make with Equiv mapDefLe.
  firstorder.
  firstorder.
  introv E1 E2. intro k. apply (PosetDec.trans _ _ _ (E1 k) (E2 k)).
  intros f g. set_eq <- b: (mapDefLeb f g). destruct b; fold_bool; rew_refl in EQb.
   left. apply~ mapDefLeb_correct.
   right. intro Le. apply EQb. apply~ mapDefLeb_correct.
Defined.

Global Instance Join : JoinDec.t t.
  apply JoinDec.Make with join; unfold join.
  (* bound1 *)
  skip. (* intros (df&mf) (dg&mg) k.
   sets_eq m: (fold (λ (k : K) (y : V), add k (dg ⊔♯ y)) mf []).
   asserts I: ((df, mf) ⊑♯ (df, m)).
     substs. clear. intro k. simpl.
     forwards E: MapFacts.fold_identity mf. sets_eq mf': (fold add mf []).
     forwards~ R: MF.fold_Equal_ord MapFacts.Equal_ST (λ k y, add k (dg ⊔♯ y)) E.
       clear. introv Ek Ev Eh. intro k. substs. apply~ MapFacts.add_m.
     rewrite <- R. rewrite <- E. rewrite EQmf'. pattern (fold add mf []).
     clear. apply MF.fold_rec. (* The invariant chosen by Coq is terribly wrong. *)
      introv E. rewrite~ fold_empty.
      introv M NI A I.
       forwards~ R: MapFacts.fold_commutes MapFacts.Equal_ST (λ k1 y, add k1 (dg ⊔♯ y)).
         clear. introv Ek Ev Eh. intro k. substs. apply~ MapFacts.add_m.
         clear. introv D. intro k''.
          tests C: (k'' === k); tests C': (k'' === k');
            repeat ((rewrite MapFacts.add_eq_o; [| solve [ auto* ] ])
                   || (rewrite MapFacts.add_neq_o; [| try unfolds; auto* ])); try reflexivity.
           false D. eapply equiv_transitive with k''; auto*.
         skip.
       rewrite R.
   clear EQm. gen I. apply MF.fold_rec. *)
  (* bound2 *)
  skip.
  (* least upper bound *)
  skip.
Defined.

Global Instance Meet : MeetDec.t t.
  apply MeetDec.Make with meet; unfold meet.
  (* bound1 *)
  skip.
  (* bound2 *)
  skip.
  (* greatest upper bound *)
  skip.
Defined.

Global Instance Lattice : LatticeDec.t t.
  eapply LatticeDec.Make.
   apply Join.
   apply Meet.
   apply BotDec.Make with (⊥♯, []).
    intros (d&m) k. simpl. rewrite~ empty_find. apply empty_1.
   (* Widen *)
   exact (fun h1 h2 => h2).
Defined.

Global Instance Top : TopDec.t t.
   apply TopDec.Make with (⊤♯, []).
   intros h k. simpl. rewrite MapFacts.empty_o. apply TopDec.prop.
Defined.

Lemma read_top : forall x, read (⊤♯) x = ⊤♯.
Proof. introv. simpls. rewrite MapFacts.empty_o. auto~. Qed.

Lemma write_monotone : forall h1 h2 k v1 v2,
  h1 ⊑♯ h2 -> v1 ⊑♯ v2 ->
  write h1 k v1 ⊑♯ write h2 k v2.
Proof.
  intros (d1&m1) (d2&m2). introv O1 O2. unfolds write.
  destruct (EquivDec.dec v1 (⊥♯)) eqn: D1; destruct (EquivDec.dec v2 (⊥♯)) eqn:D2.
   apply PosetDec.refl; auto*.
   change (⊥♯, []) with (⊥♯ : t). apply BotDec.prop.
   false n. apply~ @PosetDec.antisym. applys @PosetDec.trans O2. apply PosetDec.refl; auto*.
   intro k'. tests C: (k === k').
    do 2 eapply PosetDec.trans; try apply O2; apply PosetDec.refl; try apply EquivDec.refl.
     forwards R: read_write_eq (d1, m1) v1 C. unfolds write. rewrite D1 in R. exact R.
     forwards R: read_write_eq (d2, m2) v2 C. unfolds write. rewrite D2 in R.
      applys @EquivDec.sym R.
    do 2 eapply PosetDec.trans; try apply O1; apply PosetDec.refl; try apply EquivDec.refl.
     forwards R: read_write_neq (d1, m1) C n. unfolds write. rewrite D1 in R. exact R.
     forwards R: read_write_neq (d2, m2) C n0. unfolds write. rewrite D2 in R.
      applys @EquivDec.sym R.
Qed.

Lemma read_monotone : forall h1 h2 k,
  h1 ⊑♯ h2 ->
  read h1 k ⊑♯ read h2 k.
Proof. intros (d1&m1) (d2&m2). introv O. apply O. Qed.

Require Export LibHeap.

Variable VC : Type.
Variable gamma0 : V → VC → Prop.
Hypothesis G : Gamma.t (℘ VC) V gamma0.
Hypothesis CK : Comparable K.
Hypothesis IVC : Inhab VC.

Definition gamma (h : t) (H : heap K VC) : Prop :=
  forall k : K,
    indom H k →
    gamma0 (MapDefault.read h k) (read H k).

Global Instance Gamma : Gamma.t (℘ _) t gamma.
Proof.
  apply Gamma.Make; unfolds~ gamma.
   intros h1 h2 I H g k ID. applys~ Gamma.monotone I.
   intros h1 h2 H (M1&M2) k ID.
    forwards GC: Gamma.meet_morph (conj (M1 k ID) (M2 k ID)).
    skip. (* This proof is complex as it deals with the propagation of ⊥♯:
            M1 and M2 enforce the existence of a concrete heap H that
            respects the condition of both abstract heaps h1 and h2, and
            thus those two abstract heap were not propagated to ⊥♯ while
            merging them. *) (* TODO *)
Qed.

Lemma gamma_write : forall h H k v V,
  ~ v =♯ (⊥♯) ->
  gamma0 v V →
  gamma h H →
  gamma (MapDefault.write h k v) (write H k V).
Proof.
  introv NB Gv Gh. intros k' ID. tests C: (k' === k).
   eapply gamma_eq; [auto*| |].
    apply EquivDec.sym. apply* MapDefault.read_write_eq. apply~ @Equivalence_Symmetric.
    skip. (* Actually, [heap]s lack this quotient ability to pass through. *)
   eapply gamma_eq; [auto*| |].
    apply EquivDec.sym. apply MapDefault.read_write_neq; auto*.
    asserts: (k' ≠ k).
      intro_subst. apply C. apply Equivalence_Reflexive.
     forwards~: @indom_write_inv ID. rewrite~ read_write_neq.
Qed.

Definition gammaf (h : t) (H : K → VC) : Prop :=
  forall k : K, gamma0 (MapDefault.read h k) (H k).

Global Instance Gammaf : Gamma.t (℘ _) t gammaf.
Proof.
  apply Gamma.Make; unfolds~ gammaf.
  intros h1 h2 I H g k. applys~ @Gamma.monotone G.
  intros h1 h2 H (M1&M2) k.
   forwards: @Gamma.meet_morph G (conj (M1 k) (M2 k)).
    skip. (* This is the same… *)
Qed.

Lemma gammaf_write : forall h H H' k v,
  ~ v =♯ (⊥♯) ->
  gammaf h H →
  (forall k', ~ k === k' → H k' = H' k') →
  (forall k', k === k' → gamma0 v (H' k')) →
  gammaf (MapDefault.write h k v) H'.
Proof.
  introv NB Gh HD Gv. intros k'. tests C: (k' === k).
   eapply gamma_eq; [auto*| |].
    apply EquivDec.sym. apply* MapDefault.read_write_eq. apply~ @Equivalence_Symmetric.
    apply Gv. apply~ @Equivalence_Symmetric.
   eapply gamma_eq; [auto*| |].
    apply EquivDec.sym. apply MapDefault.read_write_neq; auto*.
    rewrite~ <- HD.
Qed.

End MapDefault.

End MapDefault.


Module PowerSet.

Section PowerSet.

Variable T : Type.
Variable all : (T → bool) → bool.
Hypothesis all_spec : forall f, all f ↔ forall t, f t.

Definition t := T → bool.

Hypothesis CT : Comparable T.

Definition const : T → t := fun t1 t2 => decide (t1 = t2).

Global Instance Equiv : EquivDec.t t.
  apply EquivDec.Make with (fun f1 f2 => forall t, f1 t = f2 t); auto*.
  intros f1 f2. sets_eq <- ra: (all (fun t => decide (f1 t = f2 t))).
  destruct ra; [left|right].
   fold_bool; rewrite all_spec in EQra. intro t. lets: (EQra t). rew_refl~ in *.
   fold_bool. rew_refl in *. rewrite all_spec in EQra. intro P; apply EQra. introv. rew_refl~.
Defined.

Global Instance Poset : PosetDec.t t.
  apply PosetDec.Make with
    (eq := Equiv) (order := fun f1 f2 : _ → bool => forall t, f1 t → f2 t).
  intros f1 f2 E t. rewrite~ E.
  intros f1 f2 I1 I2 t. lets I1': (rm I1) t. lets I2': (rm I2) t.
   destruct (f1 t); fold_bool; rew_refl in *; auto*.
  firstorder.
  intros f1 f2. set_eq <- ra: (all (fun t => decide (f1 t → f2 t))).
   destruct ra; [left|right].
    fold_bool; rewrite all_spec in EQra. intro t. lets: (EQra t). rew_refl~ in *.
    fold_bool. rew_refl in *. rewrite all_spec in EQra. intro P; apply EQra. introv. rew_refl~.
Defined.

Global Instance Lattice : LatticeDec.t t.
  apply LatticeDec.Make with Poset.
   apply JoinDec.Make with (fun (f1 f2 : T → bool) t => decide (f1 t \/ f2 t));
     try solve [ repeat intro; rew_refl* in * ].
   apply MeetDec.Make with (fun (f1 f2 : T → bool) t => decide (f1 t /\ f2 t));
     try solve [ repeat intro; rew_refl* in * ].
   apply BotDec.Make with (fun _ => false);
     try solve [ repeat intro; rew_refl* in * ].
   (* Widen *)
   exact (fun f1 f2 => f2).
Defined.

Global Instance Top : TopDec.t t.
   apply TopDec.Make with (fun _ => true);
     try solve [ repeat intro; rew_refl* in * ].
Defined.

Definition gamma (f : t) t : Prop := f t.

Global Instance Gamma : Gamma.t (℘ T) t gamma.
Proof.
  apply Gamma.Make; unfolds~ gamma.
  intros f1 f2 t M. simpls. rew_refl~.
Qed.

Lemma gamma_const : forall t t', gamma (const t) t' ↔  t = t'.
Proof. introv. unfolds gamma, const. rew_refl*. Qed.

End PowerSet.

End PowerSet.


Module SumLattice.

Section SumLattice.

Variables T1 T2 : Type.
Hypothesis L1 : LatticeDec.t T1.
Hypothesis L2 : LatticeDec.t T2.
Variables C1 C2 : Type.
Variable gamma1 : T1 → C1 → Prop.
Variable gamma2 : T2 → C2 → Prop.
Hypothesis G1 : Gamma.t (℘ C1) T1 gamma1.
Hypothesis G2 : Gamma.t (℘ C2) T2 gamma2.

Definition t0 := (T1 + T2)%type.
Definition t := option t0.

Inductive eqt0 : t0 → t0 → Prop :=
  | eqt_left : forall t1 t2,
    t1 =♯ t2 →
    eqt0 (inl t1) (inl t2)
  | eqt_right : forall t1 t2,
    t1 =♯ t2 →
    eqt0 (inr t1) (inr t2)
  | eqt_bot_rl : forall t1 t2,
    t1 =♯ (⊥♯) →
    t2 =♯ (⊥♯) →
    eqt0 (inr t1) (inl t2)
  | eqt_bot_lr : forall t1 t2,
    t1 =♯ (⊥♯) →
    t2 =♯ (⊥♯) →
    eqt0 (inl t1) (inr t2)
  .

Global Instance Equiv0 : EquivDec.t t0.
  apply EquivDec.Make with eqt0; try solve [
     repeat intros [?|?]; repeat (let H := fresh in intro H; inverts H); constructors* ].
  intros [x|x] [y|y];
    ((destruct (EquivDec.dec x y); [left; constructors*
                       |right; let H := fresh in intro H; inverts* H])
    || (destruct (EquivDec.dec x (⊥♯)); destruct (EquivDec.dec y (⊥♯));
         ((left; solve [ constructors~ ]) || (right; let H := fresh in intro H; inverts* H)))).
Defined.

Inductive order : t0 → t0 → Prop :=
  | order_left : forall t1 t2,
    t1 ⊑♯ t2 →
    order (inl t1) (inl t2)
  | order_right : forall t1 t2,
    t1 ⊑♯ t2 →
    order (inr t1) (inr t2)
  | order_bot : forall t1 t2,
    t1 =♯ inl (⊥♯) →
    order t1 t2
  .

Global Instance Poset0 : PosetDec.t t0.
  apply PosetDec.Make with
    (eq := Equiv0) (order := order).
  repeat intros [?|?]; repeat (let H := fresh in intro H; inverts H);
    (solve [ constructors~ ] || apply order_bot; constructors*).
  introv O1 O2. repeat inverts O1 as O1; repeat inverts O2 as O2; try solve [ constructors* ].
  introv O1 O2. repeat inverts O1 as O1; try solve [ apply order_bot; constructors~ ];
    repeat inverts O2 as O2; try solve [ constructors~; applys @PosetDec.trans O1 O2 ];
    apply order_bot; constructors*.
  intros [x|x] [y|y];
    (destruct (EquivDec.dec x (⊥♯)); [left; apply order_bot; constructors*|]);
    ((destruct (PosetDec.dec x y); [left; constructors~
                                   |right; intro O; repeat (inverts O as O); auto~])
     || (right; intro O; repeat inverts O as O; auto~)).
Defined.

Definition make1 t1 : t := Some (inl t1).
Definition make2 t2 : t := Some (inr t2).

Global Instance Poset : PosetDec.t t.
  apply CoLattice.Poset. apply LiftPosetDec. apply CoLattice.Poset. apply Poset0.
Defined.

Global Instance Top : TopDec.t t.
  apply TopDec.Make with None; simpl; auto*.
Defined.

Global Instance Bot : BotDec.t t.
  apply BotDec.Make with (Some (inl (⊥♯))); intros [[?|?]|]; simpl; constructors*.
Defined.

Global Instance Join : JoinDec.t t.
  apply JoinDec.Make with (fun x y =>
    match x, y return t with
    | Some (inl x), Some (inl y) => Some (inl (x ⊔♯ y))
    | Some (inr x), Some (inr y) => Some (inr (x ⊔♯ y))
    | Some (inl x), Some (inr y) =>
      ifd EquivDec.dec x (⊥♯) then Some (inr y)
      else ifd EquivDec.dec y (⊥♯) then Some (inl x)
      else None
    | Some (inr x), Some (inl y) =>
      ifd EquivDec.dec x (⊥♯) then Some (inl y)
      else ifd EquivDec.dec y (⊥♯) then Some (inr x)
      else None
    | _, _ => None
    end); repeat intros [[?|?]|];
    repeat (let H := fresh in intro H; repeat inverts H as H);
    simpls; repeat cases_if;
    try solve [ constructors; apply PosetDec.trans with (⊥♯); auto~ ];
    try solve [ repeat constructors~ ];
    constructors*.
Defined.

Global Instance Meet : MeetDec.t t.
  apply MeetDec.Make with (fun x y =>
    match x, y return t with
    | Some (inl x), Some (inl y) => Some (inl (x ⊓♯ y))
    | Some (inr x), Some (inr y) => Some (inr (x ⊓♯ y))
    | None, _ => y
    | _, None => x
    | _, _ => ⊥♯
    end); repeat intros [[?|?]|];
    repeat (let H := fresh in intro H; repeat inverts H as H); simpls;
    try solve [ constructors; apply PosetDec.trans with (⊥♯); auto~ ];
    try solve [ repeat constructors~ ].
Defined.

Global Instance Lattice : LatticeDec.t t.
  apply LatticeDec.Make with Poset; try typeclass.
Defined.

Inductive gamma : t → ℘ (C1 + C2) :=
  | gamma_left : forall c t,
    gamma1 t c →
    gamma (Some (inl t)) (inl c)
  | gamma_right : forall c t,
    gamma2 t c →
    gamma (Some (inr t)) (inr c)
  | gamma_top : forall c,
    gamma None c
  (* These are strange cases, but I didn't see where we actually assumes they can't happen. *)
  | gamma_left_bot : forall c t,
    gamma1 (⊥♯) c →
    gamma t (inl c)
  | gamma_right_bot : forall c t,
    gamma2 (⊥♯) c →
    gamma t (inr c)
  .

Global Instance Gamma : Gamma.t (℘ (C1 + C2)) t gamma.
Proof.
  apply Gamma.Make.
   intros [[t1|t1]|] [[t2|t2]|] I [c|c] G; inverts G as G; (repeat inverts I as I);
    constructors~; try applys~ @Gamma.monotone I;
    try solve [ (applys~ @Gamma.monotone G1; applys~ gamma_eq I)
             || (applys~ @Gamma.monotone G2; applys~ gamma_eq I) ].
   intros [[t1|t1]|] [[t2|t2]|] [c|c] [Hg1 Hg2]; inverts Hg1 as Hg1; inverts Hg2 as Hg2;
    constructors~; (
        (applys~ @Gamma.meet_morph G1; simpl; splits~)
     || (applys~ @Gamma.meet_morph G2; simpl; splits~) || idtac);
     (applys~ @Gamma.monotone G1 || applys~ @Gamma.monotone G2).
Qed.

End SumLattice.

End SumLattice.


Section ProductSumGamma.

Variables T1 T2 : Type.
Variable L1 : LatticeDec.t T1.
Variable L2 : LatticeDec.t T2.
Variables C1 C2 : Type.
Variable gamma1 : T1 → C1 → Prop.
Variable gamma2 : T2 → C2 → Prop.
Hypothesis G1 : Gamma.t (℘ C1) T1 gamma1.
Hypothesis G2 : Gamma.t (℘ C2) T2 gamma2.

Inductive productSumGamma : (T1 * T2) → (C1 + C2) → Prop :=
  | productSumGamma_left : forall t1 t2 c,
    gamma1 t1 c →
    productSumGamma (t1, t2) (inl c)
  | productSumGamma_right : forall t1 t2 c,
    gamma2 t2 c →
    productSumGamma (t1, t2) (inr c)
  .

Global Instance ProductSumGamma : Gamma.t (℘ (C1 + C2)) (T1 * T2) productSumGamma.
Proof.
  apply Gamma.Make.
   intros [t11 t12] [t21 t22] [I1 I2] [c|c] G; inverts G as G; simpls; constructors.
    applys~ @Gamma.monotone I1.
    applys~ @Gamma.monotone I2.
   intros [t11 t12] [t21 t22] [c|c] [Hg1 Hg2]; inverts Hg1 as Hg1; inverts Hg2 as Hg2;
     constructors; simpls.
    applys @Gamma.meet_morph G1; simpl; splits~.
    applys @Gamma.meet_morph G2; simpl; splits~.
Qed.

End ProductSumGamma.


Section ProductProdGamma.

Variables T1 T2 : Type.
Variable L1 : LatticeDec.t T1.
Variable L2 : LatticeDec.t T2.
Variables C1 C2 : Type.
Variable gamma1 : T1 → C1 → Prop.
Variable gamma2 : T2 → C2 → Prop.
Hypothesis G1 : Gamma.t (℘ C1) T1 gamma1.
Hypothesis G2 : Gamma.t (℘ C2) T2 gamma2.

Definition productProdGamma (t : T1 * T2) (c : C1 * C2) : Prop :=
  gamma1 (fst t) (fst c) ∧ gamma2 (snd t) (snd c).

Global Instance ProductProdGamma : Gamma.t (℘ (C1 * C2)) (T1 * T2) productProdGamma.
Proof.
  apply Gamma.Make.
   intros [t11 t12] [t21 t22] [I1 I2] [c1 c2] G; inverts G as G; simpls; constructors.
    applys~ @Gamma.monotone I1.
    applys~ @Gamma.monotone I2.
   intros [t11 t12] [t21 t22] [c1 c2] [Hg1 Hg2]; inverts Hg1 as Hg1; inverts Hg2 as Hg2;
     constructors; simpls.
    applys @Gamma.meet_morph G1; simpl; splits~.
    applys @Gamma.meet_morph G2; simpl; splits~.
Qed.

End ProductProdGamma.


Module Extend.

Section Extend.

Variable T : Type.
Hypothesis L : LatticeDec.t T.
Variable e : Type → Type.
Variable param : Type.
Hypothesis paramC : Comparable param.
Variable ext : forall A, param → A → e A.
Variable extr : forall A, e A → option (A * param)%type.
Hypothesis extr_ext : forall A a p, extr A (ext A p a) = Some (a, p).
Hypothesis ext_extr : forall A et a p, extr A et = Some (a, p) → ext _ p a = et.
Hypothesis extr_none_comp : forall A et1 et2,
  extr A et1 = None →
  extr A et2 = None →
  Decidable (et1 = et2).
Variable check : forall A1 A2, e A1 → e A2 → bool.
Hypothesis check_trans : forall A1 A2 A3 et1 et2 et3,
    extr A1 et1 = None → extr A2 et2 = None → extr A3 et3 = None →
  check _ _ et1 et3 → check _ _ et2 et3 → check _ _ et1 et2.
Hypothesis check_correct : forall A et1 et2,
    extr A et1 = None → extr A et2 = None →
  check  _ _ et1 et2 → et1 = et2.
Variable C : Type.
Variable gamma0 : T → C → Prop.
Hypothesis G : Gamma.t (℘ C) T gamma0.

Definition t := LiftTop.t (LiftBot.t (e T)).

Global Instance Equiv : EquivDec.t t.
  do 2 apply LiftEquivDec.
  apply EquivDec.Make with
    (eq := fun et1 et2 =>
      match extr _ et1, extr _ et2 with
      | Some (a1, p1), Some (a2, p2) => a1 =♯ a2 ∧ p1 = p2
      | _, _ => et1 = et2
      end);
    repeat (let x := fresh "x" in let a := fresh "a" in
            let p := fresh "p" in let E := fresh "E" in
            intro x; destruct (extr _ x) as [[a p]|] eqn:E);
    repeat intros (?&?); intros; substs; auto~; tryfalse.
   splits~. eapply EquivDec.trans; eassumption.
   destruct (decide (p = p0)) eqn:D; fold_bool; rew_refl in D.
    substs. destruct (EquivDec.dec a a0); [left*|right*].
    right*.
   right. intro_subst. false.
   right. intro_subst. false.
   apply decidable_sumbool. apply~ extr_none_comp.
Defined.

Global Instance Poset : PosetDec.t t.
  apply PosetDec.Make with (eq := Equiv)
    (order := fun et1 et2 =>
      match et1, et2 with
      | _, None => True
      | None, Some _ => False
      | Some eb1, Some eb2 =>
        match eb1, eb2 with
        | None, _ => True
        | Some _, None => False
        | Some t1, Some t2 =>
          match extr _ t1, extr _ t2 with
          | Some (a1, p1), Some (a2, p2) => a1 ⊑♯ a2 ∧ p1 = p2
          | None, None => t1 = t2
          | _, _ => False
          end
        end
      end);
    repeat (let x := fresh "x" in let a := fresh "a" in
            let p := fresh "p" in let E := fresh "E" in
            intros [[x|]|]; [destruct (extr _ x) as [[a p]|] eqn:E| |]);
    intros; substs; auto~; tryfalse;
    try (simpls; try rewrite E in *; try rewrite E0 in *; tryfalse; auto~);
    intuit; substs; try splits~.
   applys~ @PosetDec.trans a0.
   destruct (decide (p = p0)) eqn:D; fold_bool; rew_refl in D.
    substs. destruct (PosetDec.dec a a0); [left*|right*].
    right*.
   apply decidable_sumbool. apply~ extr_none_comp.
Defined.

Definition extr_destr A B t1 t2 C1 (C2 C3 C4 : B) :=
  match extr A t1 as sa1, extr A t2 as sa2
        return extr _ t1 = sa1 → extr _ t2 = sa2 → _ with
  | Some a1, Some a2 => fun _ _ => C1 a1 a2
  | None, None => fun E1 E2 =>
    if @decide _ (extr_none_comp _ _ _ E1 E2) then C2 else C3
  | _, _ => fun _ _ => C4
  end eq_refl eq_refl.

Lemma extr_destr_some_some : forall A B t1 t2 C1 (C2 C3 C4 : B) a1 a2,
  extr A t1 = Some a1 →
  extr A t2 = Some a2 →
  extr_destr _ _ t1 t2 C1 C2 C3 C4 = C1 a1 a2.
Proof.
  introv E1 E2. unfolds extr_destr.
  set (pi := fun E1 E2 => ifb t1 = t2 then C2 else C3). clearbody pi.
  destruct (extr A t1); tryfalse. destruct (extr A t2); tryfalse.
  inverts E1. inverts~ E2.
Qed.

Lemma extr_destr_none_none_eq : forall A B t1 C1 (C2 C3 C4 : B),
  extr A t1 = None →
  extr_destr _ _ t1 t1 C1 C2 C3 C4 = C2.
Proof.
  introv E. unfolds extr_destr.
  set (pi := fun E1 E2 => ifb t1 = t1 then C2 else C3).
  asserts E': (pi = fun _ _ => C2).
    extens. intros E1 E2. unfold pi. cases_if~.
  clearbody pi. rewrite E'. destruct~ (extr A t1); tryfalse.
Qed.

Lemma extr_destr_none_none_neq : forall A B t1 t2 C1 (C2 C3 C4 : B),
  extr A t1 = None →
  extr A t2 = None →
  t1 ≠ t2 →
  extr_destr _ _ t1 t2 C1 C2 C3 C4 = C3.
Proof.
  introv E1 E2 D. unfolds extr_destr.
  set (pi := fun E1 E2 => ifb t1 = t2 then C2 else C3).
  asserts E': (pi = fun _ _ => C3).
    extens. intros E1' E2'. unfold pi. cases_if~.
  clearbody pi. rewrite E'.
  destruct (extr A t1); tryfalse. destruct~ (extr A t2); tryfalse.
Qed.

Lemma extr_destr_some_none : forall A B t1 t2 C1 (C2 C3 C4 : B) a1,
  extr A t1 = Some a1 →
  extr A t2 = None →
  extr_destr _ _ t1 t2 C1 C2 C3 C4 = C4.
Proof.
  introv E1 E2. unfolds extr_destr.
  set (pi := fun E1 E2 => ifb t1 = t2 then C2 else C3). clearbody pi.
  destruct (extr A t1); tryfalse. destruct~ (extr A t2); tryfalse.
Qed.

Lemma extr_destr_none_some : forall A B t1 t2 C1 (C2 C3 C4 : B) a2,
  extr A t1 = None →
  extr A t2 = Some a2 →
  extr_destr _ _ t1 t2 C1 C2 C3 C4 = C4.
Proof.
  introv E1 E2. unfolds extr_destr.
  set (pi := fun E1 E2 => ifb t1 = t2 then C2 else C3). clearbody pi.
  destruct (extr A t1); tryfalse. destruct~ (extr A t2); tryfalse.
Qed.


Global Instance Lat : LatticeDec.t t.
  apply LatticeDec.Make with (porder := Poset).
   apply JoinDec.Make with
     (op := fun et1 et2 =>
       match et1, et2 with
       | Some eb1, Some eb2 =>
         match eb1, eb2 with
         | Some t1, Some t2 =>
           extr_destr _ _ t1 t2
             (fun ap1 ap2 =>
                let (a1, p1) := ap1 in let (a2, p2) := ap2 in
                ifb p1 = p2 then
                  Some (Some (ext _ p1 (a1 ⊔♯ a2)))
                else None)
             et1 None None
         | None, _ => et2
         | _, None => et1
         end
       | _, _ => None
       end);
    repeat (let x := fresh "x" in let a := fresh "a" in
            let p := fresh "p" in let E := fresh "E" in
            intros [[x|]|]; [destruct (extr _ x) as [[a p]|] eqn:E| |]);
    repeat (erewrite extr_destr_some_some; [|eassumption|eassumption]);
    repeat (erewrite extr_destr_some_none; [|eassumption|eassumption]);
    repeat (erewrite extr_destr_none_some; [|eassumption|eassumption]);
    repeat (destruct (decidable_sumbool (extr_none_comp _ _ _ E E0)) as [Cmp|Cmp];
             [ substs; rewrite~ extr_destr_none_none_eq
             | rewrite~ extr_destr_none_none_neq ]);
    intros; substs; auto~; tryfalse; simpls; try cases_if~;
    try rewrite E in *; try rewrite E0 in *; try rewrite E1 in *;
    try rewrite extr_ext; intuit; substs; auto~; tryfalse.
   apply MeetDec.Make with
     (op := fun et1 et2 =>
       match et1, et2 with
       | Some eb1, Some eb2 =>
         match eb1, eb2 with
         | Some t1, Some t2 =>
           extr_destr _ _ t1 t2
             (fun ap1 ap2 =>
                let (a1, p1) := ap1 in let (a2, p2) := ap2 in
                ifb p1 = p2 then
                  Some (Some (ext _ p1 (a1 ⊓♯ a2)))
                else Some None)
             et1 (Some None) (Some None)
         | _, _ => Some None
         end
       | None, _ => et2
       | _, None => et1
       end);
    repeat (let x := fresh "x" in let a := fresh "a" in
            let p := fresh "p" in let E := fresh "E" in
            intros [[x|]|]; [destruct (extr _ x) as [[a p]|] eqn:E| |]);
    repeat (erewrite extr_destr_some_some; [|eassumption|eassumption]);
    repeat (erewrite extr_destr_some_none; [|eassumption|eassumption]);
    repeat (erewrite extr_destr_none_some; [|eassumption|eassumption]);
    repeat (destruct (decidable_sumbool (extr_none_comp _ _ _ E E0)) as [Cmp|Cmp];
             [ substs; rewrite~ extr_destr_none_none_eq
             | rewrite~ extr_destr_none_none_neq ]);
    intros; substs; auto~; tryfalse; simpls; try cases_if~;
    try rewrite E in *; try rewrite E0 in *; try rewrite E1 in *;
    try rewrite extr_ext; intuit; substs; auto~; tryfalse.
   apply BotDec.Make with (elem := Some None). intros [?|]; simpls*.
   exact (fun x y => y).
Defined.

Global Instance Top : TopDec.t t.
  apply TopDec.Make with (elem := None). intros [?|]; simpls*.
Defined.

Definition gamma t c :=
  match t with
  | None => True
  | Some b =>
    match b with
    | None => False
    | Some e =>
      match extr _ e, extr _ c with
      | None, None => check _ _ e c
      | Some (a, ap), Some (c, cp) => ap = cp ∧ gamma0 a c
      | _, _ => False
      end
    end
  end.

Global Instance Gamma : Gamma.t (℘ (e C)) t gamma.
Proof.
  apply Gamma.Make.
   intros [[t1|]|] [[t2|]|] E c Ec; simpls~; tryfalse.
    destruct (extr _ t1) as [[a1 p1]|]; destruct (extr _ t2) as [[a2 p2]|];
      destruct (extr _ c) as [[ec ep]|]; intuit; substs~; tryfalse; splits~.
    applys~ @Gamma.monotone H1 H0.
   intros [[t1|]|] [[t2|]|] c [Hg1 Hg2]; simpls~; tryfalse.
    destruct (extr _ t1) as [[a1 p1]|] eqn:E1; destruct (extr _ t2) as [[a2 p2]|] eqn:E2;
      destruct (extr _ c) as [[ec ep]|] eqn:E; intuit; substs~; tryfalse.
     erewrite extr_destr_some_some; try eassumption. simpl. cases_if~.
      simpl. rewrite E. rewrite extr_ext. splits~. applys~ @Gamma.meet_morph G. split; auto~.
     forwards~ Hg: check_trans Hg1 Hg2. forwards~: check_correct Hg. substs.
      rewrite extr_destr_none_none_eq; try eassumption. simpl. rewrite E2. rewrite~ E.
Qed.

End Extend.

End Extend.


Section ImmerseGamma.

Variable T : Type.
Variable L : LatticeDec.t T.
Variable C : Type.
Variable gamma : T → C → Prop.
Hypothesis G : Gamma.t (℘ C) T gamma.

Variable C' : Type.
Variable f : C' → C.

Definition immerse_gamma t c := gamma t (f c).

Global Instance immerseGamma : Gamma.t (℘ C') T immerse_gamma.
Proof.
  apply Gamma.Make.
   intros t1 t2 I c g. applys @Gamma.monotone G I g.
   intros t1 t2 c [Hg1 Hg2]. applys @Gamma.meet_morph G. simpl. splits~.
Qed.

End ImmerseGamma.


